{-# LANGUAGE FlexibleContexts #-}
module Algorithms.Unify
    ( Unifiable(matchVar, eqConsts, getSubs, mapVars, apply, containsVar)
    , applyMap
    , combineMaps
    , unify
    , match
    ) where


import qualified Control.Monad.State as SM
import qualified Control.Monad.Error as EM
import qualified Data.Map as M

import Data.IdentShow


class Unifiable m where
    matchVar :: m a -> Maybe a
    eqConsts :: m a -> m b -> Bool
    getSubs :: m a -> [m a]
    mapVars :: (a -> b) -> m a -> m b
    apply :: (a -> Maybe (m a)) -> m a -> m a

    containsVar :: Eq a => a -> m a -> Bool
    containsVar x t = case matchVar t of
        Just y -> x == y
        Nothing -> any (containsVar x) (getSubs t)


applyMap :: (Ord v, Unifiable m) => M.Map v (m v) -> m v -> m v
applyMap m t = apply (\x -> M.lookup x m) t

combineMaps :: (Ord v, Unifiable m) => M.Map v (m v) -> M.Map v (m v) -> M.Map v (m v)
combineMaps m1 m2 = M.union (M.map (applyMap m2) m1) m2

data Unifiable m => MatchError m v = MatchError (m v) (m v) deriving Show

instance EM.Error (MatchError m v) where


-- match s t = sigma => sigma(s) = t
match :: (Ord k, Eq (m k), Unifiable m, EM.MonadError (MatchError m k) em) =>
         m k -> m k -> em (M.Map k (m k))
match s t = case SM.runState (EM.runErrorT $ m s t) M.empty of
                (Left err, _) -> EM.throwError err
                (Right (), subst) -> return subst
        where
            m u v = case matchVar u of
                Just x -> add x v
                Nothing -> mc u v
            mc u v =
                if eqConsts u v
                then ms (getSubs u) (getSubs v)
                else EM.throwError (MatchError u v)
                where
                    ms [] [] = return ()
                    ms (x:xs) (y:ys) = do
                        m x y
                        ms xs ys
                    ms _ _ = EM.throwError (MatchError u v)
            add x v = do
                cache <- EM.lift SM.get
                case M.lookup x cache of
                    Nothing -> EM.lift $ SM.modify (M.insert x v)
                    Just v' | v == v'   -> return ()
                            | otherwise -> EM.throwError (MatchError v v')

data Unifiable m => UnificationError m v =
      ClashFailure (m v) (m v)
    | OccurFailure v (m v)


instance (IdentShow v, Show (m v), Unifiable m) => Show (UnificationError m v) where
    show (ClashFailure l r) = "ClashFailure " ++ (show l) ++ " " ++ (show r)
    show (OccurFailure v r) = "OccurFailure " ++ (identShow v) ++ " " ++ (show r)

instance EM.Error (UnificationError m v) where

-- unify s t = sigma => sigma(s) = sigma(t)
unify :: (Eq (u v),
          Ord v,
          Unifiable u,
          EM.MonadError (UnificationError u v) em) =>
         u v -> u v -> em (M.Map v (u v))
unify s t = case SM.runState (EM.runErrorT $ u1 s t) M.empty of
                (Left err, _) -> EM.throwError err
                (Right (), subst) -> return subst
    where
        u0 u v = do
            subst <- EM.lift SM.get
            let u' = applyMap subst u
            let v' = applyMap subst v
            u1 u' v'
        u1 u v = case matchVar u of
            Just u' ->
                if matchVar v == Just u'
                then return ()
                else add u' v
            Nothing -> u2 u v
        u2 u v = case matchVar v of
            Just _ -> u1 v u
            Nothing -> u3 u v
        u3 u v =
            if eqConsts u v
            then usubs (getSubs u) (getSubs v)
            else EM.throwError (ClashFailure u v)
            where
                usubs [] [] = return ()
                usubs (x:xs) (y:ys) = do
                    u0 x y
                    usubs xs ys
                usubs _ _ = EM.throwError (ClashFailure u v)
        add x v =
            if containsVar x v
            then EM.throwError (OccurFailure x v)
            else do
                cache <- EM.lift SM.get
                case M.lookup x cache of
                    Nothing -> do
                        let f u | x == u    = Just v
                                | otherwise = Nothing
                        EM.lift $ do
                            SM.modify $ M.map (apply f)
                            SM.modify $ M.insert x v
                    Just v' | v == v'   -> return ()
                            | otherwise -> EM.throwError (ClashFailure v v')

