{-# LANGUAGE GeneralizedNewtypeDeriving #-}
module Data.Term where

import Data.Array
import qualified Data.Map as M
import qualified Data.Set as S
import qualified Data.IntSet as IS
import qualified Data.IntMap as IM

import qualified Control.Monad.State as SM

import Data.Maybe

import Data.TeX
import Data.IdentShow
import List
import Utils
import Debug
import Algorithms.Unify

data Term sym var =
      FunApp sym [Term sym var]
    | Var var 
    deriving (Eq, Ord)

type Pos = [Int]

newtype ArgVar = ArgVar Int deriving (Eq, Ord, Enum, Ix, Show)

instance IdentShow ArgVar where
    identShow (ArgVar i) = show i

instance Ord sym => Unifiable (Term sym) where
    matchVar (Var x) = Just x
    matchVar _ = Nothing
    eqConsts (FunApp f _) (FunApp g _) | f == g     = True
    eqConsts _            _                         = False
    getSubs (FunApp _ subs) = subs
    getSubs _               = []
    mapVars m (Var x) = Var (m x)
    mapVars m (FunApp f subs) = FunApp f $ map (mapVars m) subs
    apply m (Var x) = case m x of Just y -> y ; Nothing -> Var x
    apply m (FunApp f subs) = FunApp f $ map (apply m) subs


getVars :: Ord var => Term sym var -> S.Set var
getVars (Var v) = S.singleton v
getVars (FunApp _ args) = S.unions $ map getVars args

getSyms :: Ord sym => Term sym var -> S.Set sym
getSyms (Var _) = S.empty
getSyms (FunApp sym args) = S.unions $ (S.singleton sym) : (map getSyms args)





type ReplacementMap sym = M.Map sym [Bool]

type Substitution sym var = M.Map var (Term sym var)

type Path = [Int]

data Theory =
      Associative
    | Commutative
    deriving Show

data Strategy =
      Innermost
    | Outermost
    | Full
    deriving Show

data Goal
    = GComplexity

data StartTermType
    = ConstructorBased

instance (TeXable sym, TeXable var) => TeXable (Term sym var) where
    tex (FunApp f []) = "\\sym{" ++ tex f ++ "}"
    tex (FunApp f args) = "\\sym{" ++ tex f ++ "}" ++ "(" ++ (join args) ++ ")"
        where
            join [] = ""
            join [x] = tex x
            join (x:l) = tex x ++ ", " ++ join l
    tex (Var x) = tex x

data GenericRule a b
    = SimpleRule (Rule a b)
    | EquationalRule (Rule a b)
    | ConditionalRule (Rule a b) [(Rule a b)]
    deriving Show

data Rule sym var = Rule (Term sym var) (Term sym var) deriving (Eq, Ord)

getRuleVars :: Ord var => Rule sym var -> S.Set var
getRuleVars (Rule l r) = S.union (getVars l) (getVars r)

normalizeRule :: (Ord var1) => [var2] -> Rule sym var1 -> Rule sym var2
normalizeRule vars (Rule lhs rhs) = flip SM.evalState (M.empty, vars) $ do
    lhs' <- normalizeTerm lhs
    rhs' <- normalizeTerm rhs
    return $ Rule lhs' rhs'
  where
    normalizeTerm (Var v) = do
        (m, fresh:vars) <- SM.get
        case M.lookup v m of
            Just x -> return $ Var x
            Nothing -> do
                let m' = M.insert v fresh m
                SM.put (m', vars)
                return $ Var fresh
    normalizeTerm (FunApp sym subs) = mapM normalizeTerm subs >>= return . FunApp sym



type TRS sym var = [Rule sym var]

getTRSVars :: (Ord var) => TRS sym var -> S.Set var
getTRSVars = S.unions . map getRuleVars

data ContextVar sym =
      Box
    | Variable sym

type Context sym var = Term sym (ContextVar var)

data CSR sym var = CSR {
    csrStrategy :: Strategy,
    csrR :: TRS sym var,
    csrMu :: ReplacementMap sym
}

instance (IdentShow sym, IdentShow var) => Show (Term sym var) where
    show (Var x) = "$" ++ identShow x
    show (FunApp y []) = identShow y
    show (FunApp y args) = (identShow y) ++ "(" ++ (join "," $ map show args) ++ ")"
        where join :: String -> [String] -> String
              join _ [] = ""
              join _ [x] = x
              join s (x:r) = x ++ s ++ (join s r)

instance (IdentShow sym, IdentShow var) => Show (Rule sym var) where
    show (Rule l r) = show l ++ " -> " ++ show r
    showList rules s = "{\n  " ++ (concat $ intersperse "\n  " $ map show rules) ++ "\n}" ++ s

instance (IdentShow sym, IdentShow var, Ord var) => Show (CSR sym var) where
    show CSR {
        csrStrategy = strat,
        csrR = rules,
        csrMu = mu
    } = "(VARS " ++ (join " " $ map identShow $ S.toList vars) ++ ")\n"
        ++ (case strat of
                Innermost -> "(STRATEGY INNERMOST)\n"
                Outermost -> "(STRATEGY OUTERMOST)\n"
                Full -> ""
           )
        ++ "(STRATEGY CONTEXTSENSITIVE\n    "
        ++ (join "\n    " $ map showCs $ M.toList mu)
        ++ "\n)\n"
        ++ "(RULES\n    "
        ++ (join "\n    " $ map show rules)
        ++ "\n)"
        where join :: String -> [String] -> String
              join _ [] = ""
              join _ [x] = x
              join s (x:r) = x ++ s ++ (join s r)
              showCs (f, l) = "(" ++ (identShow f) ++ " "
                    ++ (join " " $ map (show . snd) $ filter fst $ zip l [1..])
                    ++  ")"
              vars = S.unions $ map getRuleVars rules

-- * Rule/Pair Array

data RuleOrPair sym var = RPRule (Rule sym var) | RPPair (Rule sym var) deriving (Eq, Ord, Show)

genRuleOrPairArray rules pairs = listArray (0, l-1) pairsAndRules
  where
    l = length pairsAndRules
    pairsAndRules = (map RPRule rules) ++ (map RPPair pairs)

genLookupMap :: (Ord sym, Ord var) => Array Int (RuleOrPair sym var) -> M.Map (RuleOrPair sym var) Int
genLookupMap = M.fromList . map (\(a,b) -> (b,a)) . assocs

-- * Obligations

-- ** QDP

data QDP sym var = QDP {
    qdpArity :: M.Map sym Int,
    qdpP :: TRS sym var,
    qdpQ :: [Term sym var],
    qdpR :: TRS sym var,
    qdpMinimal :: Bool,

    -- cached computations:

    qdpNtoRP :: Array Int (RuleOrPair sym var),
    qdpRPtoN :: M.Map (RuleOrPair sym var) Int,
    qdpDefMap :: DefMapA sym,
    qdpSymMap :: SymMap sym
}

-- ** QCSDP

data QCSDP sym var = QCSDP {
    qcsdpP :: TRS sym var,
    qcsdpQ :: [Term sym var],
    qcsdpR :: TRS sym var,
    qcsdpMu :: ReplacementMap sym,
    qcsdpMinimal :: Bool
}


symbols :: Ord sym => TRS sym var -> (S.Set sym, S.Set sym)
symbols rules = (defined, constructors)
    where defined = S.fromList $ map (\(Rule (FunApp n _) _) -> n) rules
          constructors = S.difference allSymbols defined
          allSymbols = S.unions $ concatMap (\(Rule l r) -> [getSyms l, getSyms r]) rules


arities :: Ord sym => TRS sym var -> M.Map sym Int
arities rules = SM.execState (mapM (\(Rule l r) -> do { getAr l ; getAr r }) rules) M.empty
    where getAr (Var _) = return ()
          getAr (FunApp a subs) = do
              m <- SM.get
              let l = length subs
              case M.lookup a m of
                  Nothing -> SM.modify (M.insert a l)
                  Just l'
                    | l == l' -> return ()
                    | otherwise -> fail "different arities for same symbol"
              mapM_ getAr subs


dependencies :: Ord sym => TRS sym var -> M.Map sym [sym]
dependencies rules = M.map nub deps
    where
        defSyms = fst $ symbols rules
        isDefined s = S.member s $ defSyms
        initialMap = M.fromList $ zip (S.toList defSyms) $ repeat []
        deps = SM.execState (mapM_ checkRule rules) initialMap
        checkRule (Rule (FunApp d _) r) = check r
            where check (Var _) = return ()
                  check (FunApp a subs) = do
                      mapM_ check subs
                      if isDefined a then addDep d a else return ()
        checkRule _ = error "lhs of rule is a variable"
        addDep d a = do
            m <- SM.get
            let r = assumeJust $ M.lookup d m
            SM.modify $ M.insert d (a:r)



-- a map from defined symbols to sets of rules defining this function
type DefMap sym var = M.Map sym (S.Set (Rule sym var))

createDefMap :: (Ord sym, Ord var) => TRS sym var -> DefMap sym var
createDefMap trs = SM.execState addRules M.empty
    where
        addRules = mapM_ add trs
        add rule@(Rule (FunApp f _) _) = SM.modify (M.insertWith' S.union f (S.singleton rule))

type DefMapA sym = M.Map sym IS.IntSet

createDefMapA :: (Ord sym, Ord var) => Array Int (RuleOrPair sym var) -> DefMapA sym
createDefMapA m = SM.execState addRules M.empty
    where
        addRules = mapM_ add $ assocs m
        add (k, (RPRule (Rule (FunApp f _) _))) = SM.modify (M.insertWith' IS.union f (IS.singleton k))
        add _   = return ()

-- a map from function symbols to lists of pairs and rules containing this sym
data SymMap sym = SymMap {
    symMapRules :: M.Map sym IS.IntSet,
    symMapPairs :: M.Map sym IS.IntSet
}


-- createSymMap :: (Ord sym, Ord var) => QDP sym var -> SymMap sym
createSymMap nToPR = SymMap rulemap pairmap
  where
    (rulemap, pairmap) = foldl addRP (M.empty, M.empty) $ assocs nToPR
    addRP (rm, pm) (i, (RPRule rule)) = (addRule i rule rm, pm)
    addRP (rm, pm) (i, (RPPair rule)) = (rm, addRule i rule pm)
    addRule i (Rule lhs rhs) m = foldl (add i) m $ S.toList $ S.union (getSyms lhs) (getSyms rhs)
    add i m sym = M.insertWith' IS.union sym (IS.singleton i) m



lookupRules sym symmap = case M.lookup sym symmap of
    Just rule -> rule


buildQdp :: (Ord sym, Ord var) => TRS sym var -> [Term sym var] -> TRS sym var -> Bool -> QDP sym var
buildQdp p q r minimal = QDP {
        qdpArity = arity,
        qdpP = p,
        qdpQ = q,
        qdpR = r,
        qdpMinimal = minimal,

        qdpNtoRP = ntopr,
        qdpRPtoN = genLookupMap ntopr,
        qdpDefMap = createDefMapA ntopr,
        qdpSymMap = createSymMap ntopr
    }
  where ntopr = genRuleOrPairArray r p
        arity = M.union (arities r) (arities p)

exportQDP prob =
    showVars ++
    showPairs ++
    showRules ++
    showQ ++
    showMinimal
  where
    p = qdpP prob
    r = qdpR prob
    q = qdpQ prob
    m = qdpMinimal prob
    showVars = "(VAR " ++ (concat $ intersperse " " $ map identShow vars) ++ ")\n"
    showPairs = "(PAIRS\n" ++ showList p ++ ")\n"
    showRules = "(RULES\n" ++ showList r ++ ")\n"
    showQ = "(Q\n" ++ showList q ++ ")\n"
    showMinimal = if m then "(MINIMAL)\n" else "(NOT MINIMAL)\n"
    showList r = concatMap showLine r
    showLine x = "    " ++ show x ++ "\n"
    vars = S.toList $ S.unions [getTRSVars p, getTRSVars r, qVars]
    qVars = S.unions $ map getVars q


cap prob term = SM.evalState (c term) freshVars
  where
    c (Var x) = freshVar -- before something like Q is checked, make sure that term is properly renamed before capping
    c (FunApp f subs) = do
        cappedSubs <- mapM c subs
        let newterm = FunApp f cappedSubs
        if isMatched newterm
            then freshVar
            else return newterm
    isMatched t@(FunApp s _) = any (matches t) $ rulesOf s
    matches term (Rule lhs rhs) = case unify term lhs of Left _ -> False ; Right _ -> True
    usedVars = getTRSVars $ qdpR prob
    freshVars = [ Var v | v <- map (\v -> "x_" ++ show v) [1..], not $ S.member v usedVars ]
    freshVar = do
        (v:m) <- SM.get
        SM.put m
        return v
    ithRule i = case (qdpNtoRP prob)!i of RPRule r -> r
    rulesOf sym = case M.lookup sym (qdpDefMap prob) of Just ns -> map ithRule $ IS.toList ns ; Nothing -> []


size :: Term sym var -> Int
size (Var _) = 0
size (FunApp _ args) = 1 + (sum $ map size args)
