module Data.Term.Parser where

import Data.Term
import Debug
import qualified Text.Parsec.ByteString as PBS
import Text.Parsec.Char
import Text.Parsec.Error
import Text.Parsec.Combinator
import Text.Parsec.Prim (runParser, modifyState, getState, setState, many, parserFail)
import qualified Data.ByteString as BS
import qualified Data.Set as S
import qualified Data.IntSet as IS
import qualified Data.Map as M


type STerm = Term String String
type SGenericRule = GenericRule String String

data TrsState = TrsState {
    stVars :: S.Set String,
    stArity :: M.Map String Int,
    stRules :: [SGenericRule],
    stPairs :: Maybe [SGenericRule],
    stQ :: Maybe [STerm],
    stStrategy :: Maybe Strategy,
    stMinimal :: Maybe Bool,
    stTheory :: Maybe (M.Map String Theory),
    stReplacementMap :: Maybe (M.Map String (IS.IntSet)),
    stGoal :: Maybe Goal,
    stStartTermType :: Maybe StartTermType
}

initialState = TrsState {
    stVars = S.empty,
    stArity = M.empty,
    stRules = [],
    stPairs = Nothing,
    stQ = Nothing,
    stStrategy = Nothing,
    stMinimal = Nothing,
    stTheory = Nothing,
    stReplacementMap = Nothing,
    stGoal = Nothing,
    stStartTermType = Nothing
}

addVars v s = s { stVars = S.union (S.fromList v) (stVars s) }
addRules rs s = s { stRules = rs }
addPairs rs s = s { stPairs = Just rs }
addQ q s = s { stQ = Just (oldq ++ q) }
    where oldq = case stQ s of Just x -> x ; Nothing -> []
setStrategy strat s = s { stStrategy = Just strat }
setMinimal m s = s { stMinimal = Just m }
addTheory sym t s = s { stTheory = Just $ M.insert sym t oldtheory }
    where oldtheory = case stTheory s of Just x -> x; Nothing -> M.empty
setReplacementMap rm s = s { stReplacementMap = Just rm }
setStartTermType t s = s { stStartTermType = Just t }
setGoal g s = s { stGoal = Just g }

type TrsParser = PBS.GenParser () TrsState

-- tools

checkArity sym ar = do
    s <- getState
    let am = stArity s
    case sym `M.lookup` am of
        Nothing -> do
            setState $ s { stArity = M.insert sym ar am }
        Just oldar | oldar == ar -> return ()
                   | otherwise -> parserFail
                        $ "wrong arity of symbol " ++ sym ++ ": expected " ++ (show oldar) ++ " got " ++ (show ar)


filterSimple msg (SimpleRule r) = return r
filterSimple msg _ = parserFail msg

--parsePlainTrs :: String -> BS.ByteString -> Either ParseError STRS
parsePlainTrs = runParser trsparser initialState

parseTerm vars = runParser term $ initialState { stVars = S.fromList vars }


trsparser :: TrsParser [SGenericRule]
trsparser = do
    s <- trs
    checkNothing $ stPairs s
    checkNothing $ stQ s
    checkStrategy $ stStrategy s
    checkNothing $ stMinimal s
    checkNothing $ stTheory s
    checkNothing $ stReplacementMap s
    return $ stRules s
  where
    checkNothing Nothing = return ()
    checkNothing _ = parserFail "system not supported"
    checkStrategy Nothing = return ()
    checkStrategy (Just Innermost) = return ()
    checkStrategy (Just Full) = return ()
    checkStrategy _ = parserFail "strategy not supported"


parseQDP = runParser qdpparser initialState

qdpparser :: TrsParser (QDP String String)
qdpparser = do
    s <- trs
    p <- case stPairs s of
        Just x -> return x
        Nothing -> parserFail "missing PAIRS section"
    p <- mapM (filterSimple "only simples pairs supported") p
    q <- case stQ s of
        Just x -> return x
        Nothing -> parserFail "missing Q section"
    minimal <- case stMinimal s of
        Just x -> return x
        Nothing -> parserFail "missing MINIMAL/NOT MINIMAL section"
    r <- mapM (filterSimple "only simples rules supported") $ stRules s
    let ntopr = genRuleOrPairArray r p
    return QDP {
        qdpArity = stArity s,
        qdpP = p,
        qdpQ = q,
        qdpR = r,
        qdpMinimal = minimal,
        qdpNtoRP = ntopr,
        qdpRPtoN = genLookupMap ntopr,
        qdpDefMap = createDefMapA ntopr,
        qdpSymMap = createSymMap ntopr
    }

trs = do
    spaces
    sections <- many1 section
    eof
    s <- getState
    return s

section :: TrsParser ()
section = do
    token "("
    name <- word
    case name of
        "VAR" -> var
        "RULES" -> rules
        "PAIRS" -> pairs
        "Q" -> q
        "STRATEGY" -> strategy
        "THEORY" -> theory
        "MINIMAL" -> minimal
        "NOT" -> not_minimal
        "GOAL" -> goal
        "STARTTERM" -> starttermtype
        "COMMENT" -> correctlyNested
        "Comment" -> correctlyNested
        "COMMENTS" -> correctlyNested
        "from" -> correctlyNested
        name -> parserFail $ "unknown section: " ++ name
    token ")"

-- sections
var :: TrsParser ()
var = do
    vars <- many name
    modifyState $ addVars vars

rules :: TrsParser ()
rules = do
    rules <- many rule
    modifyState $ addRules rules

pairs :: TrsParser ()
pairs = do
    pairs <- many1 rule
    modifyState $ addPairs pairs

q :: TrsParser ()
q = do
    q <- many term
    modifyState $ addQ q

minimal :: TrsParser ()
minimal = modifyState $ setMinimal True

not_minimal :: TrsParser ()
not_minimal = do
    token "MINIMAL"
    modifyState $ setMinimal False

theory :: TrsParser ()
theory = do
    many theorydecl
    return ()

theorydecl :: TrsParser ()
theorydecl = do
    token "("
    ass <- isChar 'A'
    comm <- isChar 'C'
    spaces
    symbols <- many name
    token ")"
    if ass
        then mapM_ (add Associative) symbols
        else return ()
    if comm
        then mapM_ (add Commutative) symbols
        else return ()
    return ()
  where
    isChar c = option False $ do char c; return True
    add t s = do modifyState $ addTheory s t

strategy :: TrsParser ()
strategy = do
    choice
        [ do { token "INNERMOST"; modifyState $ setStrategy Innermost },
          do { token "OUTERMOST"; modifyState $ setStrategy Outermost },
          do { token "CONTEXTSENSITIVE"; contextsensitive } ]

goal :: TrsParser ()
goal = do
    choice
        [ do { token "COMPLEXITY"; modifyState $ setGoal GComplexity } ]

starttermtype :: TrsParser ()
starttermtype = do
    choice
        [ do { token "CONSTRUCTOR-BASED"; modifyState $ setStartTermType ConstructorBased } ]

contextsensitive :: TrsParser ()
contextsensitive = do
    decls <- many csdecl
    let rm = M.fromList decls
    modifyState $ setReplacementMap rm

csdecl :: TrsParser (String, IS.IntSet)
csdecl = do
    token "("
    sym <- name
    numbers <- many int
    token ")"
    return (sym, IS.fromList numbers)

-- common stuff
rule :: TrsParser SGenericRule
rule = do
    rule <- simpleRule
    conds <- option [] $ do
        token "|"
        simpleRule `sepBy` token ","
    case (rule, conds) of
        (rule, []) -> return rule
        (EquationalRule _, _)
            -> parserFail "equational rule with condition not supported"
        (SimpleRule rule, conds) -> do
            sconds <- mapM (filterSimple "only simple rules allowed as conditions") conds
            return $ ConditionalRule rule sconds

simpleRule :: TrsParser SGenericRule
simpleRule = do
    lhs <- term
    arr <- arrow
    rhs <- term
    let rule = Rule lhs rhs
    return $ case arr of
        Rew -> SimpleRule rule
        RewEq -> EquationalRule rule

term :: TrsParser (STerm)
term = do
    f <- name
    st <- getState
    if S.member f (stVars st)
        then return (Var f)
        else do
            args <- option [] arguments
            let ar = length args
            checkArity f ar
            return (FunApp f args)

arguments :: TrsParser [STerm]
arguments = do
    token "("
    args <- term `sepBy` token ","
    token ")"
    return args

-- lexical structure

name :: TrsParser String
name = do
    n <- many1 (noneOf " \n\r\t(),")
    spaces
    return n

int :: TrsParser Int
int = do
    digits <- many1 digit
    spaces
    return $ read digits

word :: TrsParser String
word = do
    w <- many1 letter
    spaces
    return w

token :: String -> TrsParser ()
token s = do
    string s
    spaces
    return ()

correctlyNested :: TrsParser ()
correctlyNested = do
    many (noneOf "()")
    optional $ do
        char '('
        correctlyNested
        char ')'
        correctlyNested

data Arrow
    = Rew
    | RewEq

arrow :: TrsParser Arrow
arrow = do
    string "->"
    arr <- option Rew $ do
        char '='
        return RewEq
    spaces
    return arr

