module Data.Term.UsableRules where

import Data.Array
import qualified Control.Monad.State as SM
import qualified Data.Map as M
import qualified Data.Set as S
import qualified Data.IntSet as IS
import qualified Data.IntMap as IM

import Data.Term
import Algorithms.Unify
import Debug



-- * ActivePath and ActiveCondition

type ActivePathEl sym = (sym, Int)

type ActivePath sym = S.Set (ActivePathEl sym)

type ActiveRules sym var = (ActivePath sym, [Rule var sym])

type ActiveTRS sym var = [ActiveRules sym var]

type ActiveCondition sym = S.Set (ActivePath sym)

pathAnd :: (Ord sym) => ActivePath sym -> ActivePath sym -> ActivePath sym
pathAnd = S.union

pathOrCond :: (Ord sym) => ActivePath sym -> ActiveCondition sym -> ActiveCondition sym
pathOrCond path cond = newCond
  where
    paths = S.toList cond
    dropSupersets = filter (not . S.isProperSubsetOf path) paths
    containsSmaller = any (flip S.isProperSubsetOf path) paths
    newCond | containsSmaller   = cond
            | otherwise         = S.insert path $ S.fromDistinctAscList dropSupersets

pathAndCond :: (Ord sym) => ActivePath sym -> ActiveCondition sym -> ActiveCondition sym
pathAndCond path cond = S.map (pathAnd path) cond

condOr :: (Ord sym) => ActiveCondition sym -> ActiveCondition sym -> ActiveCondition sym
condOr = S.fold pathOrCond


-- * Usable Rules Estimations

type URGraph sym = IM.IntMap (IM.IntMap (ActiveCondition sym))

type UsableRules sym = [(ActivePath sym, IS.IntSet)]

combineMaps2 orig addition = IM.map updateRule orig
  where
    updateRule urmap = IM.unionsWith condOr (urmap : updated urmap)
    updated urmap = concatMap updateRule2 $ IM.toList urmap
    updateRule2 (rule, cond) = map (\path -> IM.map (pathAndCond path) reachable) paths
      where
        paths = S.toList cond
        reachable = case IM.lookup rule addition of Just x -> x


-- transitiveURClosure2 :: (Ord sym) => QDP sym var -> UsableRules sym
transitiveURClosure2 prob =
--    trace (show urgraph)
    splitConditions
  where
    urgraph = simpleDirectUsableEstimation prob
    refinements = iterate (combineMaps2 urgraph) urgraph
    fixpoint (x:y:xs)
        | x == y        = x
        | otherwise     = fixpoint (y:xs)
    graph = fixpoint refinements
    urs = IM.toList $ IM.unionsWith condOr $ map snd $ filter (isPair.fst) $ IM.toList graph
    isPair i = IS.member i pairNums
    pairNums = IS.fromDistinctAscList $ map fst $ filter pairFilter $ assocs $ qdpNtoRP prob
    pairFilter (_, RPPair _)    = True
    pairFilter _                = False
    splitConditions = M.toList $ foldl (\m (k, v) -> M.insertWith' IS.union k v m) M.empty $ concatMap split urs
    split (is, ac) = [ (ap, IS.singleton is) | ap <- S.toList ac ]



{-
simpleDirectUsableEstimation
  :: (Ord sym) =>
     QDP sym var
     -> IM.IntMap (IM.IntMap (ActiveCondition sym))
     -}
simpleDirectUsableEstimation prob = forward
  where
    forward = IM.fromList $ map (\i -> (i, directlyUsable prob S.empty $ rhs i))
        $ map fst $ assocs ropa
    ropa = qdpNtoRP prob
    ruleFilter (_, (RPRule _)) = True
    ruleFilter _ = False
    rhs i = case ropa!i of (RPRule (Rule _ rhs)) -> rhs ; (RPPair (Rule _ rhs)) -> rhs

directlyUsable prob _ (Var _) = IM.empty
directlyUsable prob p t@(FunApp f subs) = IM.unionsWith condOr (ur:subUR)
  where
    capt = cap prob t
    ithLhs i = case (qdpNtoRP prob)!i of RPRule (Rule lhs _) -> lhs
    rules = case M.lookup f (qdpDefMap prob) of Just ns -> IS.toList ns ; Nothing -> []
    matches rulenum = case unify capt (ithLhs rulenum) of Left _ -> False ; Right _ -> True
    activeRules = filter matches rules
    ur = IM.fromList $ zip activeRules (repeat $ S.singleton p)
    subPaths = map (\i -> S.insert (f, i) p) [0..]
    subUR = map (\(p, t) -> directlyUsable prob p t) $ zip subPaths subs


combineActiveMaps
  :: (Ord sym, Ord var) =>
     M.Map (Rule sym var) (M.Map (Rule sym var) (ActiveCondition sym))
     -> M.Map (Rule sym var) (M.Map (Rule sym var) (ActiveCondition sym))
     -> M.Map (Rule sym var) (M.Map (Rule sym var) (ActiveCondition sym))
combineActiveMaps orig addition = M.map updateRule orig
  where
    updateRule urmap = M.unionsWith condOr (urmap : updated urmap)
    updated urmap = concatMap updateRule2 $ M.toList urmap
    updateRule2 (rule, cond) = map (\path -> M.map (pathAndCond path) reachable) paths
      where
        paths = S.toList cond
        reachable = case M.lookup rule addition of Just x -> x

transitiveURClosure oldmap
    | oldmap == iterated    = oldmap
    | otherwise             =
--        trace "iterating UR"
        transitiveURClosure iterated
  where
    iterated = combineActiveMaps oldmap oldmap


usable
  :: (Ord sym, Ord var) =>
     [(ActivePath sym, S.Set (Rule sym var))]
     -> ActivePath sym
     -> S.Set (Rule sym var)
usable l active = S.unions $ map snd $ filter (\x -> S.isSubsetOf (fst x) active) l


usable2 l active = IS.unions $ map snd $ filter (\x -> S.isSubsetOf (fst x) active) l


-- * Graph

type NodeLabel sym var = [(ActiveCondition sym, Rule sym var)]

type EdgeLabel sym = ActiveCondition sym




