import Data.Term.Parser
import Data.Term
import Data.Maybe (catMaybes)
import System.Exit
import qualified Data.ByteString as BS
import qualified Data.Set as S
import qualified Data.Map as M
import Algorithms.Complexity.RuntimeComplexity (terminates)
import System.Environment (getArgs)
import Data.Term.TRSInfo
import System.IO


showFail :: String -> IO a
showFail message = do
    putStrLn "MAYBE"
    putStrLn message
    exitWith (ExitFailure 1)

showNoSuccess :: IO a
showNoSuccess = do
    putStrLn "MAYBE"
    exitWith ExitSuccess

showSuccess :: String -> IO a
showSuccess message = do
    putStrLn "YES(O(1),O(1))"
    putStrLn message
    exitWith ExitSuccess

data Reason
    = OnlyConstantConstructors -- no constructors have arity > 0
    | OnlyConstantDefinedSymbols -- no symbols have arity > 0
    | NoBasicApplicableRule -- no rule has only one defined symbol in the lhs
    deriving (Eq, Show)

simpleRCCheck :: Ord sym => TRSInfo sym var -> [Reason]
simpleRCCheck ti
    = catMaybes
        [ if notOnlyConstants then Nothing else Just OnlyConstantConstructors
        , if definedSymbolsNoConstants then Nothing else Just OnlyConstantDefinedSymbols
        , if hasApplicableRule then Nothing else Just NoBasicApplicableRule
        ]
  where
    def = tiDefined ti
    cons = tiConstructor ti
    positiveArity sym = case M.lookup sym $ tiArity ti of Just x -> x > 0
    notOnlyConstants = any positiveArity $ S.toList cons
    definedSymbolsNoConstants = any positiveArity $ S.toList def
    hasApplicableRule = any isApplicable $ tiTRS ti
    isApplicable (Rule (FunApp _ args) _) | all isCons args = True
                                          | otherwise = False
    isCons (FunApp sym args) = S.member sym cons && all isCons args
    isCons (Var v) = True


maxUnaryConstructors :: Ord sym => TRSInfo sym var -> Bool
maxUnaryConstructors ti = all maxUnary cons
  where
    cons = S.toList $ tiConstructor ti
    maxUnary sym = case M.lookup sym $ tiArity ti of Just x -> x <= 1


main :: IO ()
main = do
    content <- BS.getContents
    case parsePlainTrs "" content of
        Right rawtrs -> do
            if containsNonSimple rawtrs
                then return ()
                else do
                    trs <- mapM filterSimple rawtrs
                    let trsinfo = buildTRSInfo trs
                    let reasons = simpleRCCheck trsinfo
                    if NoBasicApplicableRule `elem` reasons
                        then showSuccess "All rules have at least two defined symbols in their LHS. Hence no rule is applicable to any basic term."
                        else if (OnlyConstantConstructors `elem` reasons || OnlyConstantDefinedSymbols `elem` reasons) && terminates trs
                            then  showSuccess "There are only finitely many basic terms and none of them start infinite reduction sequences."
                            else showNoSuccess
        Left error -> showFail $ show error
  where
    line l reasons = do
        putStr l
        putStr " "
        putStrLn $ show reasons
        hFlush stdout
    containsNonSimple [] = False
    containsNonSimple (SimpleRule r:rs) = containsNonSimple rs
    containsNonSimple _ = True
    filterSimple (SimpleRule r) = return r

